#include "global.h"
#include "modeman.h"
#include "zoneman.h"
#include "bddman.h"
#include "symmetry.h"
#include "print.h"

int EIM_THREAD;

int eliminate_insig(struct thread_type *);
int insignificant(struct mode_type *, struct variable_type *);
int shielded(struct mode_type *m, struct variable_type *);
int sig(struct mode_type *, struct variable_type *);
int not_in_assign(struct variable_type *, struct transition_type *);
int used_in_inv(struct variable_type *, struct mode_type *);
struct clkconj_type *given_invariance(struct mode_type *);

int reduce_predicates(struct thread_type *);
struct clkconj_type *reduce_clkconj(struct clkconj_type *);
struct variable_type *clk_out_of_bound(int, struct clkconj_type *);
int related_clkconj(struct clkconj_type *, struct variable_type *);
struct clkconj_type *delete_clkconj(struct clkconj_type **, 
                                    struct clkconj_type **);
int merge_identical_modes(struct thread_type *);
int move_out_transition(struct transition_type *, struct mode_type **, 
                        struct mode_type **);
int move_in_transition(struct transition_type *, struct mode_type **, 
                       struct mode_type **);

int add_shielded_clock(struct mode_type *, struct variable_type *);

int shield_clocks(struct thread_type *t);
int gen_used_clocks(struct thread_type *t);
struct var_link_type *get_clocks_in_trigger(struct transition_type *tr);
struct var_link_type *get_reqd_clocks(struct mode_type *m);
struct var_link_type *remove_reset_clocks(struct var_link_type *vl, 
                                          struct transition_type *tr);
struct var_link_type *delete_clock(int clk, struct var_link_type *rc);
int must_shield(struct mode_type *m, struct variable_type *v);
struct var_link_type *permuted_clocks(struct var_link_type *rc, int *perm);

int shielded_before(struct variable_type *v, struct var_link_type *sc);

/******************************************/
/* REMEMBER to annotate locations with shielded variables and do not
   include predicates of shielded variables in future compositions */
/******************************************/

int eliminate_insig(struct thread_type *t) {
  int num_var = 0;
  struct mode_type *m;
  struct variable_type *v;
  
  EIM_THREAD = t->index;

  printf("    Eliminating insignificant predicates in graph %d.\n", t->index);
  if(verbose >= 0) {
    sprintf(tmpstring, 
            "\nEliminating insignificant predicates in graph %d.\n", t->index);
    output(tmpstring);
  }
  
  for(m = t->starting_mode; m; m=m->cnext) {
    if(verbose >= 0) {
      sprintf(tmpstring, 
              "Considering mode %s (%d) with xconj_cycle:\n", 
              m->name, m->index);
      output(tmpstring);
      output("    ");
      print_clkconj(m->xconj_cycle);
      output("\n");
    }
    num_var = 0;
    for(v=var_tail->next; num_var < composed_ta;v=v->next) {
      if(v->type == V_CLOCK) {
        num_var++;
        if(insignificant(m, v) == TRUE) {
          add_shielded_clock(m, v);
          if(verbose >= 0) {
            if(v->id == NON_PARAM)
            sprintf(tmpstring, 
               "\tConjoining variable %s in mode %s (%d) with clkconj:\n\t",
                v->name, m->name, m->index);
            else
            sprintf(tmpstring, 
               "\tConjoining variable %s[%d] in mode %s (%d) with clkconj:\n\t",
                v->name, v->id, m->name, m->index);
            output(tmpstring);
            print_clkconj(m->xconj_cycle);
            output("\n");
          }
          m->xconj_cycle = conjoin(m->xconj_cycle, v->index);
          if(verbose >= 0) {
            output("\tAfter conjoining:\n\t");
            print_clkconj(m->xconj_cycle);
            output("\n");
          }
          free_zone(m->zone);
          m->zone = zone(m->xconj_cycle, t);
          m->zone = process_zone_max_const(m->zone);
          if(verbose >= 0) {
            output("\tZONE: "); 
            print_zone(m->zone); 
            output("\n");
          }
        }
      }
      if(v == var_tail) break;
    }
  }
  printf("\tMerging identical modes in graph %d.\n", t->index);
  if(verbose >= 0) {
    sprintf(tmpstring, "\nMerging identical modes in graph %d.\n", t->index);
    output(tmpstring);
  }

  merge_identical_modes(t);
  print_counts(t);

  printf("\tReducing multigraph of graph %d.\n", t->index);
  reduce_multigraph(t);
  if(verbose >= 0) {
    sprintf(tmpstring, "DONE reducing multigraph of graph %d.\n", t->index);
    output(tmpstring);
  }
  if(SGM_MODE != GUI && SGM_MODE != PROCESS) {
    fprintf(stderr, "\t");
    print_counts(t);
  }
  return 0;
}

int insignificant(struct mode_type *m, struct variable_type *v) {
  struct mode_type *p;
  for(p = thread[EIM_THREAD].starting_mode; p; p=p->cnext) {
    p->status = UNVISITED;
  }
  return shielded(m, v);
}

int shielded(struct mode_type *m, struct variable_type *v) {
  struct transition_type *tr;
  struct clkconj_type *c;

  if(verbose >= 1) {
    sprintf(tmpstring, 
            "Checking if clock %s[%d] must be shielded in mode %d(%s)\n",  
            v->name, v->id, m->index, m->name);
    output(tmpstring);
  }

  /* (1) Checking visited */
  if(m->status == VISITED) {
    if(verbose >= 1) {
      sprintf(tmpstring, "    Mode %d(%s) is visited.\n", m->index, m->name);
      output(tmpstring);
    }
    return TRUE;
  }
  m->status = VISITED;

  if(m->out_trans_cycle) {
    for(tr=m->out_trans_cycle->next; ; tr=tr->next) {

      /* (2) Check transition clock triggers */
      if(tr->trn->clk_trigger_cycle) {
        if(tr->trn->clk_trigger_cycle->op == TRUE) {  /* no trigger on tr */
          if(tr == m->out_trans_cycle) break;
          else continue;
        }
        for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
          if(c->clk1 == v->index || c->clk2 == v->index) {
            if(verbose >= 1) {
              sprintf(tmpstring,
   "    Clock %s[%d] is used in trigger of transition %d from mode %s (%d).\n",
                v->name, v->id, tr->index, m->name, m->index);
              output(tmpstring);
            }
            return FALSE;
          }
          if(c == tr->trn->clk_trigger_cycle) break;
        }
      }

      /* (3) Check reset assignments */
      if(tr->trn->assign_cycle) {
        if(not_in_assign(v, tr) == FALSE) { /* clock v is reset on tr*/
          if(verbose >= 1) {
            sprintf(tmpstring, "    Clock %s[%d] is reset on transition %d.\n",
                    v->name, v->index, tr->index);
            output(tmpstring);
          }
          if(tr == m->out_trans_cycle) break;
          else continue;
        }
      }

      /* (4) Recurse with out-mode */
      if( shielded(tr->mode, 
                   var_ptr_from_index(permute_variable(v->index, 
                                                       tr->permutation->perm)) 
                  )
          == FALSE) {
        if(verbose >= 1) {
          sprintf(tmpstring, "Clock %s[%d] is never tested or reset.\n", 
                  v->name, v->id);
          output(tmpstring);
        }
        return FALSE;
      }
      else
        if(tr == m->out_trans_cycle) break;
    }
  }
  if(verbose >= 1) {
    sprintf(tmpstring, 
            "Clock %s[%d] is never tested or reset.\n", v->name, v->id);
    output(tmpstring);
  }
  return TRUE;
}

int add_shielded_clock(struct mode_type *m, struct variable_type *v) {
  struct var_link_type *vl = (struct var_link_type *)
                             malloc(sizeof(struct var_link_type));;
  vl->var = v;
  m->shielded_clocks = insert_var_link(vl, m->shielded_clocks);
  return 0;
}



/* SECOND: this is the second ad-hoc method of determining clock significance
 *         from the mode vector of modes in the Fischer's Mutual Exclusion
 *         example
 */

/*
int sig_clk_from_mode_vector(struct mode_type *m, struct variable_type *v) {
  int i, num_auto;
  char **n = (char **)malloc(sizeof(char *) * (system_ta_num + 1));
  char *tc = (char *)malloc(sizeof(char) * (strlen(m->name) + 1));
  char *tmp_name = (char *)malloc(sizeof(char)*(strlen(m->name)+1));
  char *new_name = (char *)malloc(sizeof(char)*(strlen(m->name)+1));

  tmp_name = strcpy(tmp_name, m->name);
  tc = strtok(tmp_name, ".");

  n[1] = (char *)malloc(sizeof(char) * (strlen(tc) + 1));
  n[1] = strcpy(n[1], tc);

  i = 1;
  while(tc = strtok((char *)NULL, ".")) {
    i++;
    n[i] = (char *)malloc(sizeof(char) * (strlen(tc) + 1));
    n[i] = strcpy(n[i], tc);
  }
  num_auto = i;

  if(!strcmp(n[v->id], "M1") || !strcmp(n[v->id], "M4")) return FALSE;
  if(!strcmp(n[v->id], "M2")) return TRUE;
  for(i=1; i<=num_auto; i++) {
    if(!strcmp(n[i], "M2")) {
      return FALSE;
    }
  }
  return TRUE;
}
*/

/* FIRST: this is the original clock significance indicator function */

int sig(struct mode_type *m, struct variable_type *v) {
  struct transition_type *tr;
  struct clkconj_type *c;

  if(m->status == VISITED) return FALSE;
  m->status = VISITED;

  /* (1) check if used in the given invariance of mode m 
   */
  if(used_in_inv(v, m)) {
    if(verbose >= 0) {
      sprintf(tmpstring, "    Variable %s[%d] is used in mode %s (%d).\n",
              v->name, v->id, m->name, m->index);
      output(tmpstring);
    }
    return TRUE;
  }
  
  if(m->out_trans_cycle) {
    /* (2) check if variable v is used in any trigger of out-going transitions
     *     of mode m 
     */
    for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
      if(tr->trn->clk_trigger_cycle) {
        if(tr->trn->clk_trigger_cycle->op == TRUE) {
          if(tr == m->out_trans_cycle) break;
          else continue;
        }
        for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
          if(c->clk1 == v->index || c->clk2 == v->index) {
            if(verbose >= 0) {
              sprintf(tmpstring, 
              "    Variable %s[%d] is used in trans-trigger of mode %s (%d).\n",
                v->name, v->id, m->name, m->index);
              output(tmpstring);
            }
            return TRUE; 
          }
          if(c == tr->trn->clk_trigger_cycle) break;
        } 
      }
      if(tr == m->out_trans_cycle) break;
    }

    /* (3) if v is not used in assignment of out-trans of mode m,
     *     check sig of out-mode 
     */
    for(tr=m->out_trans_cycle->next; ; tr=tr->next) {
      if(not_in_assign(v, tr)) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
           "    Variable %s[%d] is not used in assignment of mode %s (%d).\n", 
           v->name, v->id, m->name, m->index);
          output(tmpstring);
          sprintf(tmpstring, 
           "    Further checking the out-mode %s(%d) by trans %d with ",
            tr->mode->name, tr->mode->index, tr->index);
          output(tmpstring);
          print_permutation(tr->permutation);
        }
        if(tr->permutation) {
          if(sig(tr->mode, 
               var_ptr_from_index(perm_index(v->index, tr->permutation->perm)) 
              ) == TRUE)
          return TRUE;
        }
        else if(sig(tr->mode, v) == TRUE) return TRUE;
      }
      if(tr == m->out_trans_cycle) break;
    }
  }
  return FALSE;
}

int not_in_assign(struct variable_type *v, struct transition_type *tr) {
  struct assign_type *a;
  if(tr->trn->assign_cycle) {
    for(a = tr->trn->assign_cycle->next; ; a=a->next) {
      if(a->var == v->index) return FALSE;
      if(a == tr->trn->assign_cycle) break;
    }
  }
  return TRUE;
}

int used_in_inv(struct variable_type *v, struct mode_type *m) {
  struct clkconj_type *p, *g;
  
  g = given_invariance(m);
  if(g) {
    for(p = g->next; ; p = p->next) {
      if(p->clk1 == v->index) return TRUE;
      if(p->clk2 == v->index) return TRUE;
      if(p == g) break;
    }
  } 
  return FALSE;
}

struct clkconj_type *given_invariance(struct mode_type *m) {
  struct clkconj_type *ans = NULL, *p;
  struct mode_type *mptr;
  char *tmp_name = str_preserve(m->name);
  char *tc = strtok(tmp_name, ".");

  while(tc) {
    for(mptr = thread[0].starting_mode; mptr; mptr=mptr->cnext) {
      if(!strcmp(mptr->name, tc)) {
        p = clkconj_and(mptr->xconj_cycle, ans);
        free_clkconj(&ans);
        ans = p;
        break;
      }
    }
    tc = strtok((char *)NULL, ".");
  }
/*
  sprintf(tmpstring, 
          "Given invariance of mode %s (%d) is:\n", m->name, m->index);
  output(tmpstring);
  print_clkconj(ans);
  output("\n");
*/
  return ans;
}

int reduce_predicates(struct thread_type *t) {
  struct mode_type *m;

  printf("Reducing predicates in graph %d.\n", t->index);
  sprintf(tmpstring, "\nReducing predicates in graph %d.\n", t->index);
  output(tmpstring);

  for(m = t->starting_mode; m; m=m->cnext) {
    if(verbose >= 0) {
      sprintf(tmpstring, 
              "Considering mode %s (%d) with xconj_cycle:\n", 
              m->name, m->index);
      output(tmpstring);
      output("    ");
      print_clkconj(m->xconj_cycle);
      output("\n");
    }
    m->xconj_cycle = reduce_clkconj(m->xconj_cycle);
    free_zone(m->zone);
    output("Freed zone.\n");
    m->zone = zone(m->xconj_cycle, t);
    output("Assigned new zone:\n");
    print_zone(m->zone);
    output("\n");
  }
  print_counts(t);

  printf("\tMerging identical modes in graph %d.\n", t->index);
  sprintf(tmpstring, "\nMerging identical modes in graph %d.\n", t->index);
  output(tmpstring);

  merge_identical_modes(t);
  print_counts(t);
  return 0;
}

struct clkconj_type *reduce_clkconj(struct clkconj_type *c) {
  struct clkconj_type *pptr, *ptr;
  struct variable_type *v;
  int i;

  if(c) {
    for(i = 1; i<=total_clocks; i++) {
      if( (v = clk_out_of_bound(i, c)) != NULL) {
        if(verbose >= 0) {
          sprintf(tmpstring, 
                  "    Clock %s[%d] is out of bounds.\n", v->name, v->id);
          output(tmpstring);
        }
        for(pptr = c, ptr = c->next; ;) {
          if(related_clkconj(ptr, v)) {
            if(ptr == c) {
              c = pptr;
              output("\tRelated clkconj: ");
              print_one_clkconj(ptr);
              delete_clkconj(&ptr, &c);
              output("\tAfter deleting related clkconj:\n\t");
              print_clkconj(c);
              output("\n");
              break;
            }
            output("\tRelated clkconj: ");
            print_one_clkconj(ptr);
            delete_clkconj(&ptr, &c);
            output("\tAfter deleting related clkconj:\n\t");
            print_clkconj(c);
            output("\n");
            ptr = pptr->next;
          }
          else {
            if(ptr == c) break;
            pptr = ptr;
            ptr = ptr->next;
          }
        }
      }
    }
  }
  return c;
}

struct variable_type *clk_out_of_bound(int clk, struct clkconj_type *c) {
  struct variable_type *v;
  struct clkconj_type *tmp_clkconj, *cc;
 
  if(var_tail) {
    for(v=var_tail->next; ;v=v->next) {
      if(v->type == V_CLOCK) {
        if(v->id == clk) {
          tmp_clkconj = alloc_clkconj();
          tmp_clkconj->clk1 = v->index;
          tmp_clkconj->op = LEQ;
          tmp_clkconj->diff = max_const;
          cc = clkconj_and(tmp_clkconj, c);
          free_clkconj(&tmp_clkconj);
          if(cc->op == FALSE) {
            free_clkconj(&cc);
            return v;
          }
          else {
            free_clkconj(&cc);
            return (struct variable_type *)NULL;
          }
        }
      }
    }
  }
  return NULL;
}

int related_clkconj(struct clkconj_type *c, struct variable_type *v) {
  if(c->clk1 == v->index) {
    if(c->clk2 != 0) return TRUE;
    if(c->op == GREATER && c->diff >= max_const) 
      return FALSE;    /* this clkconj is "v > c_max" itself */
    return TRUE;
  }
  if(c->clk2 == v->index) {
    if(c->clk1 != 0) return TRUE;
    if(c->op == LESS && c->diff <= -max_const)
      return FALSE;    /* this clkconj is "-v < -c_max" itself */
    return TRUE;
  }
  return FALSE;
}

struct clkconj_type *delete_clkconj(struct clkconj_type **c, 
                                    struct clkconj_type **cycle) {
  struct clkconj_type *pp, *p;

  if((*cycle)->next == *cycle && *c == *cycle) {
    free_clkconj(cycle);
    *cycle = NULL;
    return *cycle;
  }

  if(*cycle) {
    for(pp = *cycle, p = (*cycle)->next; ;) {
      if(p == *c) {
        if(p == *cycle) {
          *cycle = pp;
          pp->next = p->next;
          (*c)->next = *c;
          free_clkconj(c);
          break;
        }
        else {
          pp->next = p->next;
          p = p->next;
          (*c)->next = *c;
          free_clkconj(c);
        }
      }
      else {
        if(p == *cycle) break;
        pp = p;
        p = p->next;
      }
    }
  }
  return *cycle;
}

int merge_identical_modes(struct thread_type *t) {
  int i;
  struct mode_type *m, *n, *np = NULL;
  struct transition_type *tr, *trs;

  for(i=1, m=t->starting_mode; m; m=m->cnext, i++) {
    if(m->cnext)
    for(np=m, n=m->cnext; n;) {
      if((m == n) || (mode_comp(m, n) != EQUAL)) {
        np = n;
        n = n->cnext;
        continue;
      }
      if(verbose >= 0) {
        output("Merging identical modes: ");
/*
        sprintf(tmpstring, 
                "%s(%d) [%s(%d), %s(%d)] and %s(%d) [%s(%d), %s(%d)].\n",
                m->name, m->index, 
                m->submode1->name, m->submode1->index, 
                m->submode2->name, m->submode2->index,
                n->name, n->index, 
                n->submode1->name, n->submode1->index,
                n->submode2->name, n->submode2->index);
*/
        sprintf(tmpstring, "%s(%d) and %s(%d).\n",
                           m->name, m->index, n->name, n->index);
        output(tmpstring);
      }
      if(n->out_trans_cycle) {
        for(tr=n->out_trans_cycle->next, trs=tr->next; 
            n->out_trans_count > 0; tr=trs, trs=trs->next) {
          move_out_transition(tr, &n, &m);
        }
      }
      if(n->in_trans_cycle) {
        for(tr=n->in_trans_cycle->next, trs=tr->next; 
            n->in_trans_count > 0; tr=trs, trs=trs->next) {
          move_in_transition(tr, &n, &m);
        }
      }
      if(np == NULL) {
        t->starting_mode = n->cnext;
        free_mode(n);
        n = t->starting_mode;
        t->mode_count--;
      }
      else {
        np->cnext = n->cnext;
        free_mode(n);
        t->mode_count--;
        n = np->cnext;
      }
    }
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "\tChecked %d modes, Remaining %d modes (%.1f%%)\r", 
                      i, t->mode_count-i, 
              ((float)(t->mode_count-i) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();
  if(SGM_MODE != GUI && SGM_MODE != PROCESS)
    fprintf(stderr, "\t");
  return 0;
}


/* move an out-going transition of mode n from mode n to mode m */

int move_out_transition(struct transition_type *tr, struct mode_type **n, 
                        struct mode_type **m) {
  struct transition_type *pp, *p;

  if(verbose >= 2) {
    sprintf(tmpstring, "Moving out-trans %d from mode %s(%d) to mode %s(%d).\n",
           tr->index, (*n)->name, (*n)->index, (*m)->name, (*m)->index);
    output(tmpstring);
  }
  if(tr->mode->in_trans_cycle) {
    for(p = tr->mode->in_trans_cycle->next; ; p = p->next) {
      if(p->index == tr->index) {
        p->mode = (*m);
        break;
      }
/*
      if(p == tr->mode->in_trans_cycle) break;
*/
    }
  }
  if((*n)->out_trans_cycle) {
    (*n)->out_trans_count--;
    if((*n)->out_trans_cycle == (*n)->out_trans_cycle->next &&
       tr->index == (*n)->out_trans_cycle->index) 
      (*n)->out_trans_cycle = NULL;
    else 
    for(pp = (*n)->out_trans_cycle, p=pp->next; ; pp=p, p=p->next) {
      if(p->index == tr->index) {
        pp->next = tr->next;
        if((*n)->out_trans_cycle == p)
          (*n)->out_trans_cycle = pp;
        break;
      }
/*
      if(p == (*n)->out_trans_cycle) break;
*/
    }
  }
  (*m)->out_trans_count++;
  if((*m)->out_trans_cycle) {
    tr->next = (*m)->out_trans_cycle->next;
    (*m)->out_trans_cycle->next = tr;
    (*m)->out_trans_cycle = tr;
  }
  else {
    tr->next = tr;
    (*m)->out_trans_cycle = tr;
  }
  return 0;
}

/* Function found NOT required -----
 * After moving a transition from mode n to mode m,
 * update the transition permutation into
 * compose(inverse(m_perm), n_perm);
 */
/*
int update_moved_trans_perm(struct transition_type *tr, struct mode_type *n,
                            struct mode_type *m) {
  int *pptr;
  struct permutation_type *p = alloc_permutation();
  pptr = inverse_permutation(m->norm_perm->perm);
  p->perm = compose_permutation(pptr, n->norm_perm->perm);
  free(pptr);
  pptr = p->perm;
  if(tr->permutation) {
    p->perm = compose_permutation(pptr, tr->permutation->perm);
    free(pptr);
  } 

  tr->permutation = (struct permutation_type *)
                     add_23tree(p, perm_tree, permutation_comp,
                     free_perm, perm_count_inc, perm_parent_set,
                     print_permutation);
  return 0;
}
*/


/* move an in-coming transition of mode n from mode n to mode m */

int move_in_transition(struct transition_type *tr, struct mode_type **n, 
                       struct mode_type **m) {
  struct transition_type *pp, *p;

  if(verbose >= 2) {
    sprintf(tmpstring, "Moving in-trans %d from mode %s(%d) to mode %s(%d).\n",
           tr->index, (*n)->name, (*n)->index, (*m)->name, (*m)->index);
    output(tmpstring);
  }
  if(tr->mode->out_trans_cycle) {
    for(p = tr->mode->out_trans_cycle->next; ; p = p->next) {
      if(p->index == tr->index) {
        p->mode = (*m);
        break;
      }
/*
      if(p == tr->mode->out_trans_cycle) break;
*/
    }
  }
  if((*n)->in_trans_cycle) {
    (*n)->in_trans_count--;
    if((*n)->in_trans_cycle == (*n)->in_trans_cycle->next &&
       tr->index == (*n)->in_trans_cycle->index) 
      (*n)->in_trans_cycle = NULL;
    else
    for(pp = (*n)->in_trans_cycle, p=pp->next; ; pp=p, p=p->next) {
      if(p->index == tr->index) {
        pp->next = tr->next;
        if((*n)->in_trans_cycle == p)
          (*n)->in_trans_cycle = pp;
        break;
      }
/*
      if(p == (*n)->in_trans_cycle) break;
*/
    }
  }
  (*m)->in_trans_count++;
  if((*m)->in_trans_cycle) {
    tr->next = (*m)->in_trans_cycle->next;
    (*m)->in_trans_cycle->next = tr;
    (*m)->in_trans_cycle = tr;
  }
  else {
    tr->next = tr;
    (*m)->in_trans_cycle = tr;
  }
  return 0;
}

int shield_clocks(struct thread_type *t) {
/*
  int num_var = 0;
*/
  int i;
  struct mode_type *m;
  struct variable_type *v;
  char *c = get_graph_name(t);
  
  if(t->mode_count == 0) {
    fprintf(stderr, "Thread %s contains no mode.\n", c);
    return 1;
  }

  EIM_THREAD = t->index;
  t->mnpl |= SC;
  NEW_THREAD = t;

  printf("    Shielding clocks in graph %s.\n", c);
  fflush(stdout);
  if(verbose >= 0) {
    sprintf(tmpstring, "\nShielding clocks in graph %s.\n", c);
    output(tmpstring);
  }
  
  gen_used_clocks(t);

  if(verbose >= 0)
    print_clock_indexes(t->starting_mode->zone);

  for(m = t->starting_mode, i=1; m; m=m->cnext,i++) {
    if(verbose >= 1) {
      sprintf(tmpstring, 
              "Considering mode %s (%d) with xconj_cycle:\n", 
              m->name, m->index);
      output(tmpstring);
      output("    ");
      print_clkconj(m->xconj_cycle);
      output("\n");
    }
/*
    num_var = 0;
*/
    for(v=var_tail->next; ;v=v->next) {
      if(v->type == V_CLOCK && access_clock(v, t)) {
/*
        num_var++;
*/
        if(must_shield(m, v) == TRUE) {

/*
          sprintf(tmpstring, "(Before) Shielded Clocks of Mode %s = ", m->name);
          output(tmpstring);
          print_var_links(m->shielded_clocks);
*/

          add_shielded_clock(m, v);

/*
          sprintf(tmpstring, "(After) Shielded Clocks of Mode %s = ", m->name);
          output(tmpstring);
          print_var_links(m->shielded_clocks);
*/

          if(verbose >= 1) {
            if(v->id == NON_PARAM) 
            sprintf(tmpstring, 
               "\tConjoining variable %s in mode %s (%d) with clkconj:\n\t",
                v->name, m->name, m->index);
            else 
            sprintf(tmpstring, 
               "\tConjoining variable %s[%d] in mode %s (%d) with clkconj:\n\t",
                v->name, v->id, m->name, m->index);
            output(tmpstring);
            print_clkconj(m->xconj_cycle);
            output("\n");
          }
          m->xconj_cycle = conjoin(m->xconj_cycle, v->index);
          if(verbose >= 1) {
            output("\tAfter conjoining:\n\t");
            print_clkconj(m->xconj_cycle);
            output("\n");
          }
          free_zone(m->zone);
          m->zone = zone(m->xconj_cycle, t);
          m->zone = process_zone_max_const(m->zone);
          if(verbose >= 1) {
            output("\tZONE: "); 
            print_zone(m->zone); 
            output("\n");
          }
        }
      }
      if(v == var_tail) break;
    }
    if(stats == TRUE && SGM_MODE != GUI && SGM_MODE != PROCESS) {
      clear_line();
      fprintf(stderr, "    SC-Checked %d modes, Remaining %d modes (%.1f%%)\r",
                      i, t->mode_count-i,
              ((float)(t->mode_count-i) / (float)t->mode_count)*100.0);
    }
  }
  clean_line();
  merge_iden_red_multi(t);
  free(c);
  return 0;
}

int gen_used_clocks(struct thread_type *t) {
  int flag = TRUE, *tmp_perm1;
  struct mode_type *m;
  struct transition_type *tr;
  struct var_link_type *vl, *vl_tmp, *vl_tmp2;

  for(m = t->starting_mode; m; m=m->cnext) {
    m->reqd_clocks = get_reqd_clocks(m);
  }
  while(flag) {
    if(verbose >= 0)
      output("\n\n");
    flag = FALSE;
    for(m = t->starting_mode; m; m=m->cnext) {
      if(m->out_trans_cycle) {
        for(tr = m->out_trans_cycle->next; ; tr=tr->next) {
          if(verbose >= 2) {
            sprintf(tmpstring, 
                    "Before transition %d, mode %s(%d) has required clocks = ", 
                    tr->index, m->name, m->index);
            output(tmpstring);
            print_var_links(m->reqd_clocks);
          }

          if(tr->permutation)
            tmp_perm1 = inverse_permutation(tr->permutation->perm);
          else 
            tmp_perm1 = NULL;
          if(tmp_perm1 != NULL) {
            vl = permuted_clocks(tr->mode->reqd_clocks, tmp_perm1);
            vl = sort_var_links(vl);
            free(tmp_perm1);
          }
          else 
            vl = var_link_copy(tr->mode->reqd_clocks);

          if(verbose >= 2) {
            sprintf(tmpstring, 
                    "Destination mode %s(%d) of transition %d needs clocks: ", 
                    tr->mode->name, tr->mode->index, tr->index);
            output(tmpstring);
            print_var_links(vl);
          }

          vl = remove_reset_clocks(vl, tr);
          if(verbose >= 2) {
            output("After removing reset clocks, needed clocks = ");
            print_var_links(vl);
          }
          vl_tmp = clocks_union(m->reqd_clocks, vl);
          free_var_links(vl);
          vl = vl_tmp;

          vl_tmp = get_clocks_in_trigger(tr);

          if(verbose >= 2) {
            sprintf(tmpstring, "Tr %d needs clocks: ", tr->index);
            output(tmpstring);
            print_var_links(vl_tmp);
          }

          vl_tmp2 = clocks_union(vl, vl_tmp);
          free_var_links(vl_tmp);
          free_var_links(vl);
          vl = vl_tmp2;

          if(shield_clocks_compare(vl, m->reqd_clocks) != EQUAL)
            flag = TRUE;
          free_var_links(m->reqd_clocks);
          m->reqd_clocks = vl;
          if(verbose >= 2) {
            sprintf(tmpstring, 
                    "After transition %d, mode %s(%d) has required clocks = ", 
                    tr->index, m->name, m->index);
            output(tmpstring);
            print_var_links(m->reqd_clocks);
          }
          if(tr == m->out_trans_cycle) break;
        }
      }
    }
  }
  return 0;
}

int get_thread_index(int v) {
  int i;
  for(i=0; i<system_ta_num; i++) {
    if(access_clock(var_ptr_from_index(v), &thread[i])) 
      return thread[i].index;
  }
  fprintf(stderr, "No graph found which access clock %s[%d].\n",
                  var_name(v), var_id(v));
  return -1;
}

struct var_link_type *get_reqd_clocks(struct mode_type *m) {
  char *tc; /* = (char *)malloc(sizeof(char) * (strlen(m->name) + 1)); */
  char *tmp_name = (char *)malloc(sizeof(char)*(strlen(m->name)+1));
  struct var_link_type *vl, *ans=NULL;
  struct mode_type *mm;
  struct clkconj_type *c;
  struct transition_type *tr;
  int i=1, ti;

  tmp_name = strcpy(tmp_name, m->name);
  tc = strtok(tmp_name, ".");
 
  /* check all clocks used in invariance conds of modes in mode-vector of m */
/*
  while(TRUE) {
    if(tc == NULL) break;
    ti = get_thread_index(m->zone->clk_index[i]);
*/
  for(i=0; i<system_ta_num; i++) {
   if(m->mode_vec[i]) {
    ti = i;
    tc = m->mode_vec[i];
    for(mm=thread[ti].starting_mode; mm; mm=mm->cnext) {
      if(!strcmp(mm->name, tc)) {
        if(mm->xconj_cycle) {
          if(mm->xconj_cycle->op == TRUE) continue;
          for(c = mm->xconj_cycle->next; ; c=c->next) {
            if(shielded_before(var_ptr_from_index(c->clk1), m->shielded_clocks) 
               == FALSE) {
              vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
              vl->var = var_ptr_from_index(c->clk1);
              ans = insert_var_link(vl, ans);
            }
            if(c == mm->xconj_cycle) break;
          }
        }
      }
    }
/*
    tc = strtok((char *)NULL, ".");
    i++;
*/
   }
  }
  free(tmp_name);
  /* check all clocks used in trigger of out-going transitions of m */
  if(m->out_trans_cycle) {
    for(tr = m->out_trans_cycle->next; ; tr=tr->next) {
      if(tr->trn->clk_trigger_cycle) {
        if(tr->trn->clk_trigger_cycle->op == TRUE) {
          if(tr == m->out_trans_cycle) break;
          else continue;
        }
        for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
          if(shielded_before(var_ptr_from_index(c->clk1), m->shielded_clocks) 
             == FALSE) {
            vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
            vl->var = var_ptr_from_index(c->clk1);
            ans = insert_var_link(vl, ans);
          }
          if(c == tr->trn->clk_trigger_cycle) break;
        }
      }
      if(tr == m->out_trans_cycle) break;
    }
  }
  return ans;
}

struct var_link_type *get_clocks_in_trigger(struct transition_type *tr) {
  struct clkconj_type *c;
  struct var_link_type *vl, *ans = NULL;

  if(tr->trn->clk_trigger_cycle) {
    if(tr->trn->clk_trigger_cycle->op == TRUE)
      return ans;
    for(c = tr->trn->clk_trigger_cycle->next; ; c=c->next) {
      vl = (struct var_link_type *)malloc(sizeof(struct var_link_type));
      vl->var = var_ptr_from_index(c->clk1);
      ans = insert_var_link(vl, ans);
      if(c == tr->trn->clk_trigger_cycle) break;
    }
  }
  return ans;
}

struct var_link_type *remove_reset_clocks(struct var_link_type *vl, 
                                          struct transition_type *tr) {
  struct assign_type *a;
  if(tr->trn->assign_cycle) {
    for(a = tr->trn->assign_cycle->next; ; a=a->next) {
      if(get_var_type(a->var) == V_CLOCK) {
        vl = delete_clock(a->var, vl);
      } 
      if(a == tr->trn->assign_cycle) break;
    }
  }
  return vl;
}

struct var_link_type *delete_clock(int clk, struct var_link_type *rc) {
  struct var_link_type *vlp, *vl;
  
  if(rc) {
    if(rc->next == rc) {
      if(rc->var->index == clk) {
        free_var_links(rc); rc = NULL;
      }
      return rc;
    }
    for(vlp = rc, vl=vlp->next; ; vlp=vl, vl=vl->next) {
      if(vl->var->index == clk) {
        vlp->next = vl->next;
        vl->next = vl;
        if(vl == rc) rc = vlp;
        free_var_links(vl);
        break;
      }
      if(vl == rc) break;
    }
  }
  return rc;
}

int must_shield(struct mode_type *m, struct variable_type *v) {
  struct var_link_type *vl;
  if(m->reqd_clocks) {
    for(vl = m->reqd_clocks->next; ; vl=vl->next) {
      if(vl->var->index == v->index) return FALSE;
      if(vl == m->reqd_clocks) break;
    }
  }
  return TRUE;
}

struct var_link_type *permuted_clocks(struct var_link_type *rc, int *perm) {
  struct var_link_type *rcn = NULL, *vl;

  if(rc) {
    rcn = var_link_copy(rc);
    for(vl = rcn->next; ; vl=vl->next) {
      vl->var = var_ptr_from_index(permute_variable(vl->var->index, perm));
      if(vl == rcn) break;
    }
  }
  return rcn;
}


/* Check if clock v was shielded in the past */

int shielded_before(struct variable_type *v, struct var_link_type *sc) {
  struct var_link_type *vl;
  if(sc) {
    for(vl=sc->next; ; vl=vl->next) {
      if(vl->var->index == v->index) return TRUE;
      if(vl == sc) break;
    }
  }
  return FALSE;
}

